"""
This is the primary class for user interaction with the catalog
"""

from __future__ import annotations
from typing import Optional
import os
import json
import glob
from copy import deepcopy
import logging

from pyArango.connection import Connection
from pyArango.database import Database
from pyArango.document import Document

import pandas as pd
import numpy as np

from astropy.coordinates import SkyCoord, search_around_sky
from astropy.table import Table
from astropy import units as u

from .transient import Transient
from ..exceptions import FailedQueryError, OtterLimitationError, TransientMergeError
from ..util import bibcode_to_hrn, freq_to_obstype, freq_to_band, _DuplicateFilter

import warnings

warnings.simplefilter("once", RuntimeWarning)
warnings.simplefilter("once", UserWarning)
warnings.simplefilter("once", u.UnitsWarning)

logger = logging.getLogger(__name__)


def _np_encoder(object):
    """
    Numpy data type encoder for json.dump
    """
    if isinstance(object, (np.generic, np.ndarray)):
        return object.item()


class Otter(Database):
    """
    This is the primary class for users to access the otter backend database

    Args:
        url (str): The url where the database api endpoints are located
        username (str): The username to log into the database with
        password (str): The password to log into the database with
        gen_summary (bool): Generate a local summary table, this should generally be
                            left as False!
        datadir (str): Path to the data directory with the otter data. If not provided
                       will default to a ".otter" directory in the CWD where you call
                       this class from.
        debug (bool): If we should just debug and not do anything serious.

    Returns:
        An Otter object that is connected to the otter database
    """

    def __init__(
        self,
        url: str = "http://127.0.0.1:8529",
        username: str = os.environ.get("ARANGO_USER_USERNAME", "user-guest"),
        password: str = os.environ.get("ARANGO_USER_PASSWORD", ""),
        gen_summary: bool = False,
        datadir: str = None,
        debug: bool = False,
        **kwargs,
    ) -> None:
        print("Attempting to login with the following credentials:")
        print(f"username: {username}")
        print(f"password: {password}")

        # save inputs
        if datadir is None:
            self.CWD = os.path.dirname(os.path.abspath("__FILE__"))
            self.DATADIR = os.path.join(self.CWD, ".otter")
        else:
            self.CWD = os.path.dirname(datadir)
            self.DATADIR = datadir

        self.debug = debug

        if gen_summary:
            self.generate_summary_table(save=True)

        # make sure the data directory exists
        if not os.path.exists(self.DATADIR):
            try:
                os.makedirs(self.DATADIR)
            except FileExistsError:
                logger.warning(
                    "Directory was created between the if statement and trying "
                    + "to create the directory!"
                )
                pass

        connection = Connection(username=username, password=password, arangoURL=url)
        super().__init__(connection, "otter", **kwargs)

    def get_meta(self, **kwargs) -> Table:
        """
        Get the metadata of the objects matching the arguments

        Args:
            **kwargs : Arguments to pass to Otter.query(). See that documentation with
                       `help(otter.Otter.query)`.
        Return:
           The metadata for the transients that match the arguments. Will be an astropy
           Table by default, if raw=True will be a dictionary.
        """
        metakeys = [
            "name",
            "coordinate",
            "date_reference",
            "distance",
            "classification",
            "reference_alias",
        ]

        return [t[metakeys] for t in self.query(**kwargs)]

    def cone_search(
        self, coords: SkyCoord, radius: float = 5, raw: bool = False
    ) -> Table:
        """
        Performs a cone search of the catalog over the given coords and radius.

        Args:
            coords (SkyCoord): An astropy SkyCoord object with coordinates to match to
            radius (float): The radius of the cone in arcseconds, default is 0.05"
            raw (bool): If False (the default) return an astropy table of the metadata
                        for matching objects. Otherwise, return the raw json dicts

        Return:
            The metadata for the transients in coords+radius. Will return an astropy
            Table if raw is False, otherwise a dict.
        """

        transients = self.query(coords=coords, radius=radius, raw=raw)

        return transients

    def get_phot(
        self,
        flux_unit="mag(AB)",
        date_unit="MJD",
        return_type="astropy",
        obs_type=None,
        keep_raw=False,
        wave_unit="nm",
        freq_unit="GHz",
        **kwargs,
    ) -> Table:
        """
        Get the photometry of the objects matching the arguments. This will do the
        unit conversion for you!

        Args:
            flux_unit (astropy.unit.Unit): Either a valid string to convert
                                            or an astropy.unit.Unit
            date_unit (astropy.unit.Unit): Either a valid string to convert to a date
                                            or an astropy.unit.Unit
            return_type (str): Either 'astropy' or 'pandas'. If astropy, returns an
                               astropy Table. If pandas, returns a pandas DataFrame.
                               Default is 'astropy'.
            obs_type (str): Either 'radio', 'uvoir', or 'xray'. Will only return that
                            type of photometry if not None. Default is None and will
                            return any type of photometry.
            keep_raw (bool): If True, keep the raw flux/date/freq/wave associated with
                             the dataset. Else, just keep the converted data. Default
                             is False.
            wave_unit (str): The astropy wavelength unit to return with
            freq_unit (str): The astropy frequency unit to return with`
            **kwargs : Arguments to pass to Otter.query(). Can be::

                       names (list[str]): A list of names to get the metadata for
                       coords (SkyCoord): An astropy SkyCoord object with coordinates
                                          to match to
                       radius (float): The radius in arcseconds for a cone search,
                                       default is 0.05"
                       minZ (float): The minimum redshift to search for
                       maxZ (float): The maximum redshift to search for
                       refs (list[str]): A list of ads bibcodes to match to. Will only
                                         return metadata for transients that have this
                                         as a reference.
                       hasSpec (bool): if True, only return events that have spectra.

        Return:
            The photometry for the requested transients that match the arguments.
            Will be an astropy Table sorted by transient default name.

        Raises:
            FailedQueryError: When the query returns no results
            IOError: if one of your inputs is incorrect
        """
        warn_filt = _DuplicateFilter()
        logger.addFilter(warn_filt)

        queryres = self.query(hasphot=True, **kwargs)

        dicts = []
        for transient in queryres:
            # clean the photometry
            default_name = transient["name/default_name"]

            try:
                phot = transient.clean_photometry(
                    flux_unit=flux_unit,
                    date_unit=date_unit,
                    wave_unit=wave_unit,
                    freq_unit=freq_unit,
                    obs_type=obs_type,
                )

                phot["name"] = [default_name] * len(phot)

                dicts.append(phot)

            except FailedQueryError:
                # This is fine, it just means that there is no data associated
                # with this one transient. We'll check and make sure there is data
                # associated with at least one of the transients later!
                pass

        if len(dicts) == 0:
            raise FailedQueryError()
        fullphot = pd.concat(dicts)

        # remove some possibly confusing keys
        keys_to_keep = [
            "name",
            "converted_flux",
            "converted_flux_err",
            "converted_date",
            "converted_wave",
            "converted_freq",
            "converted_flux_unit",
            "converted_date_unit",
            "converted_wave_unit",
            "converted_freq_unit",
            "filter_name",
            "obs_type",
            "upperlimit",
            "reference",
            "human_readable_refs",
        ]

        if "upperlimit" not in fullphot:
            fullphot["upperlimit"] = False

        if not keep_raw:
            if "telescope" in fullphot:
                fullphot = fullphot[keys_to_keep + ["telescope"]]
            else:
                fullphot = fullphot[keys_to_keep]

        logger.removeFilter(warn_filt)
        if return_type == "astropy":
            return Table.from_pandas(fullphot)
        elif return_type == "pandas":
            return fullphot
        else:
            raise IOError("return_type can only be pandas or astropy")

    def load_file(self, filename: str) -> dict:
        """
        Loads an otter JSON file

        Args:
            filename (str): The path to the OTTER JSON file to load

        Returns:
            dictionary with the otter JSON file contents
        """

        # read in files from summary
        with open(filename, "r") as f:
            to_ret = Transient(json.load(f))

        return to_ret

    def query(
        self,
        names: list[str] = None,
        coords: SkyCoord = None,
        radius: float = 5,
        minz: float = None,
        maxz: float = None,
        refs: list[str] = None,
        hasphot: bool = False,
        hasspec: bool = False,
        spec_classed: bool = False,
        unambiguous: bool = False,
        classification: str = None,
        class_confidence_threshold: float = 0,
        query_private=False,
        **kwargs,
    ) -> dict:
        """
        Searches the arango database table and reads relevant JSON files

        WARNING! This does not do any conversions for you!
        This is how it differs from the `get_meta` method. Users should prefer to use
        `get_meta`, and `get_phot` independently because it is a better
        workflow and can return the data in an astropy table with everything in the
        same units.

        Args:
            names (list[str]): A list of names to get the metadata for
            coords (SkyCoord): An astropy SkyCoord object with coordinates to match to
            radius (float): The radius in arcseconds for a cone search, default is 0.05"
            minz (float): The minimum redshift to search for
            maxz (float): The maximum redshift to search for
            refs (list[str]): A list of ads bibcodes to match to. Will only return
                              metadata for transients that have this as a reference.
            hasphot (bool): if True, only returns transients which have photometry.
            hasspec (bool): NOT IMPLEMENTED! Will return False for all targets!
            spec_classed (bool): If True, only returns transients that have been
                                 specotroscopically classified/confirmed
            unambiguous (bool): If True, only returns transients that only have a single
                                published classification in OTTER. If classifications
                                disagree for a transient, it will be filtered out.
            classification (str): A classification string to search for
            class_confidence_threshold (float): classification confidence cutoff for
                                                query, between 0 and 1. Default is 0.
            query_private (bool): Set to True if you would like to also query the
                                  dataset located at whatever you set datadir to

        Return:
           Get all of the raw (unconverted!) data for objects that match the criteria.
        """
        # write some AQL filters based on the inputs
        query_filters = ""

        if hasphot is True:
            query_filters += "FILTER 'photometry' IN ATTRIBUTES(transient)\n"

        if hasspec is True:
            query_filters += "FILTER 'spectra' IN ATTRIBUTES(transient)\n"

        if spec_classed:
            query_filters += "FILTER transient.classification.spec_classed >= 1"

        if unambiguous:
            query_filters += "FILTER transient.classification.unambiguous"

        if classification is not None:
            query_filters += f"""
            FOR subdoc IN transient.classification.value
                FILTER subdoc.confidence > TO_NUMBER({class_confidence_threshold})
                FILTER subdoc.object_class LIKE '%{classification}%'
            """

        if minz is not None:
            sfilt = f"""
            FILTER 'redshift' IN transient.distance[*].distance_type
            LET redshifts1 = (
                FOR val IN transient.distance
                FILTER val.distance_type == 'redshift'
                FILTER TO_NUMBER(val.value) >= {minz}
                RETURN val
            )
            FILTER COUNT(redshifts1) > 0
            """
            query_filters += sfilt
        if maxz is not None:
            sfilt = f"""
            FILTER 'redshift' IN transient.distance[*].distance_type
            LET redshifts2 = (
                FOR val IN transient.distance
                FILTER val.distance_type == 'redshift'
                FILTER TO_NUMBER(val.value) <= {maxz}
                RETURN val
            )
            FILTER COUNT(redshifts2) > 0
            """
            query_filters += sfilt

        if names is not None:
            if isinstance(names, str):
                query_filters += f"""
            FILTER UPPER(transient.name) LIKE UPPER('%{names}%')\n
                """
            elif isinstance(names, list):
                namefilt = f"""
            FOR name IN {names}
                FILTER name IN transient.name.alias[*].value\n
                """
                query_filters += namefilt
            else:
                raise Exception("Names must be either a string or list")

        if refs is not None:
            if isinstance(refs, str):  # this is just a single bibcode
                query_filters += f"FILTER {refs} IN transient.reference_alias[*].name"
            elif isinstance(refs, list):
                query_filters += f"""
                FOR ref IN {refs}
                    FILTER ref IN transient.reference_alias[*].name
                """
            else:
                raise Exception("reference list must be either a string or a list")

        # define the query
        query = f"""
        FOR transient IN transients
            {query_filters}
            RETURN transient
        """

        # set batch size to 100 million (for now at least)
        result = self.AQLQuery(query, rawResults=True, batchSize=100_000_000)

        # now that we have the query results do the RA and Dec queries if they exist
        if coords is not None:
            # get the catalog RAs and Decs to compare against
            query_coords = coords
            good_tdes = []

            for tde in result:
                for coordinfo in tde["coordinate"]:
                    if "ra" in coordinfo and "dec" in coordinfo:
                        coord = SkyCoord(
                            coordinfo["ra"],
                            coordinfo["dec"],
                            unit=(coordinfo["ra_units"], coordinfo["dec_units"]),
                        )
                    elif "l" in coordinfo and "b" in coordinfo:
                        # this is galactic
                        coord = SkyCoord(
                            coordinfo["l"],
                            coordinfo["b"],
                            unit=(coordinfo["l_units"], coordinfo["b_units"]),
                            frame="galactic",
                        )
                    else:
                        raise ValueError(
                            "Either needs to have ra and dec or l and b as keys!"
                        )
                    if query_coords.separation(coord) < radius * u.arcsec:
                        good_tdes.append(tde)
                        break  # we've confirmed this tde is in the cone!

            arango_query_results = [Transient(t) for t in good_tdes]

        else:
            arango_query_results = [Transient(res) for res in result.result]

        if not query_private:
            return arango_query_results

        private_results = self._query_datadir(
            names=names,
            coords=coords,
            radius=radius,
            minz=minz,
            maxz=maxz,
            refs=refs,
            hasphot=hasphot,
            hasspec=hasspec,
        )

        partially_merged = deepcopy(arango_query_results)
        new_transients = []
        for jj, t_private in enumerate(private_results):
            for ii, t_public in enumerate(arango_query_results):
                try:
                    partially_merged[ii] += t_private
                    break
                except TransientMergeError:
                    continue
            else:
                new_transients.append(t_private)

        return partially_merged + new_transients

    def _query_datadir(
        self,
        names: list[str] = None,
        coords: SkyCoord = None,
        radius: float = 5,
        minz: float = None,
        maxz: float = None,
        refs: list[str] = None,
        hasphot: bool = False,
        hasspec: bool = False,
        raw: bool = False,
    ) -> dict:
        """
        This is a private method and is here just for the pipeline!!!
        This should not be used by end users!

        Searches the summary.csv table and reads relevant JSON files

        WARNING! This does not do any conversions for you!
        This is how it differs from the `get_meta` method. Users should prefer to use
        `get_meta`, `getPhot`, and `getSpec` independently because it is a better
        workflow and can return the data in an astropy table with everything in the
        same units.

        Args:
            names (list[str]): A list of names to get the metadata for
            coords (SkyCoord): An astropy SkyCoord object with coordinates to match to
            radius (float): The radius in arcseconds for a cone search, default is 0.05"
            minz (float): The minimum redshift to search for
            maxz (float): The maximum redshift to search for
            refs (list[str]): A list of ads bibcodes to match to. Will only return
                              metadata for transients that have this as a reference.
            hasphot (bool): if True, only returns transients which have photometry.
            hasspec (bool): if True, only return transients that have spectra.

        Return:
           Get all of the raw (unconverted!) data for objects that match the criteria.
        """
        if (
            all(arg is None for arg in [names, coords, maxz, minz, refs])
            and not hasphot
            and not hasspec
        ):
            # there's nothing to query!
            # read in the metdata from all json files
            # this could be dangerous later on!!
            allfiles = glob.glob(os.path.join(self.DATADIR, "*.json"))
            jsondata = [self.load_file(jsonfile) for jsonfile in allfiles]

            return jsondata

        # check if the summary table exists, if it doen't create it
        summary_table = os.path.join(self.DATADIR, "summary.csv")
        if not os.path.exists(summary_table):
            self.generate_summary_table(save=True)

        # then read and query the summary table
        summary = pd.read_csv(summary_table)
        if len(summary) == 0:
            return []

        # coordinate search first
        if coords is not None:
            if not isinstance(coords, SkyCoord):
                raise ValueError("Input coordinate must be an astropy SkyCoord!")
            summary_coords = SkyCoord(
                summary.ra.tolist(), summary.dec.tolist(), unit=(u.deg, u.deg)
            )

            try:
                summary_idx, _, _, _ = search_around_sky(
                    summary_coords, coords, seplimit=radius * u.arcsec
                )
            except ValueError:
                summary_idx, _, _, _ = search_around_sky(
                    summary_coords,
                    SkyCoord([coords]),
                    seplimit=radius * u.arcsec,
                )

            summary = summary.iloc[summary_idx]

        # redshift
        if minz is not None:
            summary = summary[summary.z.astype(float) >= minz]

        if maxz is not None:
            summary = summary[summary.z.astype(float) <= maxz]

        # check photometry and spectra
        if hasphot:
            summary = summary[summary.hasPhot == True]

        if hasspec:
            summary = summary[summary.hasSpec == True]

        # check names
        if names is not None:
            if isinstance(names, str):
                n = {names}
            else:
                n = set(names)

            checknames = []
            for alias_row in summary.alias:
                rs = set(eval(alias_row))
                intersection = list(n & rs)
                checknames.append(len(intersection) > 0)

            summary = summary[checknames]

        # check references
        if refs is not None:
            checkrefs = []

            if isinstance(refs, str):
                n = {refs}
            else:
                n = set(refs)

            for ref_row in summary.refs:
                rs = set(eval(ref_row))
                intersection = list(n & rs)
                checkrefs.append(len(intersection) > 0)

            summary = summary[checkrefs]

        outdata = [self.load_file(path) for path in summary.json_path]

        return outdata

    def upload(self, json_data, collection="vetting", testing=False) -> Document:
        """
        Upload json_data to collection WITHOUT deduplication!

        Args:
            json_data [dict] : A dictionary of the json data to upload to Otter
            collection [str] : The collection to upload to
            testing [bool] : Default is False

        Returns:
            The pyArango document that was uplaoded
        """

        # now add the document
        doc = self[collection].createDocument(json_data)
        if not testing:
            doc.save()
        return doc

    def upload_private(self, collection="vetting", testing=False) -> None:
        """
        Upload the local/private data stored in self.DATADIR to the vetting collection
        (like a SQL table) in the central arangodb document database.

        WARNING! This will make any data in self.DATADIR public! Please double check
        before using this method!

        Args:
            collection (str) : The collection to add the documents to. Default is
                               "vetting" where the documents will then be vetted by
                               our team.
            testing (bool) : True if don't actually upload, False is default

        Returns:
            If testing is false (the default), returns the arangodb upload result. If
            testing is true, returns the list of merged dictionaries that would get
            uploaded.

        Raises:
            OtterLimitationError: If some objects in OTTER are within 5" we can't figure
                                  out which ones to merge with which ones.

        """

        if not self.hasCollection(collection):
            raise ValueError(f"{collection} not in {self}!")

        local_data = self._query_datadir()
        docs = []
        for t in local_data:
            res = self.query(coords=t.get_skycoord())

            if len(res) > 1:
                raise OtterLimitationError("Some objects in Otter are too close!")

            elif len(res) == 1:
                # this object exists in otter already, let's grab the transient data and
                # merge the files
                merged = t + res[0]

                # copy over the special arangodb keys
                merged["_key"] = res[0]["_key"]
                merged["_id"] = res[0]["_id"]

                # we also have to delete the document from the OTTER database
                doc = self.fetchDocument(merged["_id"])
                if not testing:
                    doc.delete()
                else:
                    print(f"Would delete\n{doc}")

            else:
                # this means the object doesn't exist in otter already
                merged = t

            docs.append(self.upload(merged, collection=collection, testing=testing))

        return docs

    def save(self, schema: list[dict], testing=False) -> None:
        """
        Upload all the data in the given list of schemas.

        Args:
            schema (list[dict]): A list of json dictionaries
            testing (bool): Should we just enter test mode? Default is False

        Raises:
            OtterLimitationError: If some objects in OTTER are within 5" we can't figure
                                  out which ones to merge with which ones.
        """

        if not isinstance(schema, list):
            schema = [schema]

        for transient in schema:
            # convert the json to a Transient
            if not isinstance(transient, Transient):
                transient = Transient(transient)

            print(transient["name/default_name"])

            coord = transient.get_skycoord()
            res = self._query_datadir(coords=coord)

            if len(res) == 0:
                # This is a new object to upload
                print("Adding this as a new object...")
                self._save_document(dict(transient), test_mode=testing)

            else:
                # We must merge this with existing data
                print("Found this object in the database already, merging the data...")
                if len(res) == 1:
                    # we can just add these to merge them!
                    combined = res[0] + transient
                    self._save_document(combined, test_mode=testing)
                else:
                    # for now throw an error
                    # this is a limitation we can come back to fix if it is causing
                    # problems though!
                    raise OtterLimitationError("Some objects in Otter are too close!")

        # update the summary table appropriately
        self.generate_summary_table(save=True)

    def _save_document(self, schema, test_mode=False):
        """
        Save a json file in the correct format to the OTTER data directory
        """
        # check if this documents key is in the database already
        # and if so remove it!
        jsonpath = os.path.join(self.DATADIR, "*.json")
        aliases = {item["value"].replace(" ", "-") for item in schema["name"]["alias"]}
        filenames = {
            os.path.basename(fname).split(".")[0] for fname in glob.glob(jsonpath)
        }
        todel = list(aliases & filenames)

        # now save this data
        # create a new file in self.DATADIR with this
        if len(todel) > 0:
            outfilepath = os.path.join(self.DATADIR, todel[0] + ".json")
            if test_mode:
                print("Renaming the following file for backups: ", outfilepath)
        else:
            if test_mode:
                print("Don't need to mess with the files at all!")
            fname = schema["name"]["default_name"] + ".json"
            fname = fname.replace(" ", "-")  # replace spaces in the filename
            outfilepath = os.path.join(self.DATADIR, fname)

        # format as a json
        if isinstance(schema, Transient):
            schema = dict(schema)

        out = json.dumps(schema, indent=4, default=_np_encoder)
        # out = '[' + out
        # out += ']'

        if not test_mode:
            with open(outfilepath, "w") as f:
                f.write(out)
        else:
            print(f"Would write to {outfilepath}")
            # print(out)

    def generate_summary_table(self, save=False) -> pd.DataFrame:
        """
        Generate a summary table for the JSON files in self.DATADIR

        args:
            save (bool): if True, save the summary file to "summary.csv"
                         in self.DATADIR. Default is False and is just returned.

        returns:
            pandas.DataFrame of the summary meta information of the transients
        """
        allfiles = glob.glob(os.path.join(self.DATADIR, "*.json"))

        # read the data from all the json files and convert to Transients
        rows = []
        for jsonfile in allfiles:
            with open(jsonfile, "r") as j:
                t = Transient(json.load(j))
                skycoord = t.get_skycoord()

                row = {
                    "name": t.default_name,
                    "alias": [alias["value"] for alias in t["name"]["alias"]],
                    "ra": skycoord.ra,
                    "dec": skycoord.dec,
                    "refs": [ref["name"] for ref in t["reference_alias"]],
                }

                if "date_reference" in t:
                    date_types = {d["date_type"] for d in t["date_reference"]}
                    if "discovery" in date_types:
                        row["discovery_date"] = t.get_discovery_date()

                if "distance" in t:
                    dist_types = {d["distance_type"] for d in t["distance"]}
                    if "redshift" in dist_types:
                        row["z"] = t.get_redshift()

                row["hasPhot"] = "photometry" in t
                row["hasSpec"] = "spectra" in t

                row["json_path"] = os.path.abspath(jsonfile)

                rows.append(row)

        alljsons = pd.DataFrame(rows)
        if save:
            alljsons.to_csv(os.path.join(self.DATADIR, "summary.csv"))

        return alljsons

    @staticmethod
    def from_csvs(
        metafile: str,
        photfile: str = None,
        local_outpath: Optional[str] = None,
        db: Otter = None,
    ) -> Otter:
        """
        Converts private metadata and photometry csvs to an Otter object stored
        *locally* so you don't need to worry about accidentally uploading them to the
        real Otter database.

        Args:
            metafile (str) : String filepath or string io csv object of the csv metadata
            photfile (str) : String filepath or string io csv object of the csv
                                          photometry
            local_outpath (str) : The outpath to write the OTTER json files to
            db (Otter) : An Otter instance to add the local_outpath to for querying.
                         This keyword can be useful if you have special permission for
                         the otter database and want to upload your private data

        Returns:
            An Otter object where the json files are stored locally
        """
        # read in the metadata and photometry file
        meta = pd.read_csv(metafile)
        meta.columns = meta.columns.str.strip()  # clean up the col names
        phot = None

        required_phot_cols = [
            "name",
            "date",
            "date_format",
            "filter",
            "filter_eff",
            "filter_eff_units",
            "flux",
            "flux_err",
            "flux_unit",
        ]

        if photfile is not None:
            phot_unclean = pd.read_csv(photfile)
            phot_unclean.columns = phot_unclean.columns.str.strip()  # cleanup colnames

            phot = phot_unclean.dropna(subset=required_phot_cols)
            if len(phot) != len(phot_unclean):
                logger.warning("""
                Filtered out rows with nan in the photometry file! Make sure you
                expect this behaviour!
                """)

            if "bibcode" not in phot:
                phot["bibcode"] = "private"
                logger.warning("""
                Setting the bibcode column to the special keyword 'private'!
                """)

            # we need to generate columns of wave_eff and freq_eff
            wave_eff = []
            freq_eff = []
            wave_eff_unit = u.nm
            freq_eff_unit = u.GHz
            for val, unit in zip(phot.filter_eff, phot.filter_eff_units):
                wave_eff.append(
                    (val * u.Unit(unit))
                    .to(wave_eff_unit, equivalencies=u.spectral())
                    .value
                )
                freq_eff.append(
                    (val * u.Unit(unit))
                    .to(freq_eff_unit, equivalencies=u.spectral())
                    .value
                )

            phot["band_eff_wave"] = wave_eff
            phot["band_eff_wave_unit"] = str(wave_eff_unit)
            phot["band_eff_freq"] = freq_eff
            phot["band_eff_freq_unit"] = str(freq_eff_unit)

        if not os.path.exists(local_outpath):
            os.makedirs(local_outpath)

        # drop duplicated names in meta and keep the first
        meta = meta.drop_duplicates(subset="name", keep="first")

        # merge the meta and phot data
        if phot is not None:
            data = pd.merge(phot, meta, on="name", how="inner")
        else:
            data = meta

        # perform some data checks
        assert (
            len(data[pd.isna(data.ra)].name.unique()) == 0
        ), "Missing some RA and Decs, please check the input files!"
        if phot is not None:
            for name in meta.name:
                assert len(data[data.name == name]) == len(
                    phot[phot.name == name]
                ), f"failed on {name}"

        # actually do the data conversion to OTTER
        all_jsons = []
        for name, tde in data.groupby("name"):
            json = {}
            tde = tde.reset_index()

            # name first
            json["name"] = dict(
                default_name=name,
                alias=[dict(value=name, reference=[tde.coord_bibcode[0]])],
            )

            # coordinates
            json["coordinate"] = [
                dict(
                    ra=tde.ra[0],
                    dec=tde.dec[0],
                    ra_units=tde.ra_unit[0],
                    dec_units=tde.dec_unit[0],
                    reference=[tde.coord_bibcode[0]],
                    coordinate_type="equatorial",
                )
            ]

            ### distance info
            json["distance"] = []

            # redshift
            if "redshift" in tde and not np.any(pd.isna(tde["redshift"])):
                json["distance"].append(
                    dict(
                        value=tde.redshift[0],
                        reference=[tde.redshift_bibcode[0]],
                        computed=False,
                        distance_type="redshift",
                    )
                )

            # luminosity distance
            if "luminosity_distance" in tde and not np.any(
                pd.isna(tde["luminosity_distance"])
            ):
                json["distance"].append(
                    value=tde.luminosity_distance[0],
                    reference=[tde.luminosity_distance_bibcode[0]],
                    unit=tde.luminosity_distance_unit[0],
                    computed=False,
                    distance_type="luminosity",
                )

            # comoving distance
            if "comoving_distance" in tde and not np.any(
                pd.isna(tde["comoving_distance"])
            ):
                json["distance"].append(
                    value=tde.comoving_distance[0],
                    reference=[tde.comoving_distance_bibcode[0]],
                    unit=tde.comoving_distance_unit[0],
                    computed=False,
                    distance_type="comoving",
                )

            # remove the distance list if it is empty still
            if len(json["distance"]) == 0:
                del json["distance"]

            ### Classification information that is in the csvs
            # classification
            if "classification" in tde:
                class_flag = 0
                if "classification_flag" in tde:
                    class_flag = tde.classification_flag[0]
                json["classification"] = dict(
                    value=[
                        dict(
                            object_class=tde.classification[0],
                            confidence=class_flag,
                            reference=[tde.classification_bibcode[0]],
                        )
                    ]
                )

            # discovery date
            # print(tde)
            if "discovery_date" in tde and not np.any(pd.isna(tde.discovery_date)):
                json["date_reference"] = [
                    dict(
                        value=str(tde.discovery_date.tolist()[0]).strip(),
                        date_format=tde.discovery_date_format.tolist()[0].lower(),
                        reference=tde.discovery_date_bibcode.tolist(),
                        computed=False,
                        date_type="discovery",
                    )
                ]

            # host information
            if "host_ref" in tde and not np.any(pd.isna(tde.host_ref)):
                host_info = dict(
                    host_name=tde.host_name.tolist()[0].strip(),
                    host_ra=tde.host_ra.tolist()[0],
                    host_dec=tde.host_dec.tolist()[0],
                    host_ra_units=tde.host_ra_unit.tolist()[0],
                    host_dec_units=tde.host_dec_unit.tolist()[0],
                    reference=[tde.host_ref.tolist()[0]],
                )

                if not pd.isna(tde.host_redshift.tolist()[0]):
                    host_info["host_z"] = tde.host_redshift.tolist()[0]

                if "host" in json:
                    json["host"].append(host_info)
                else:
                    json["host"] = [host_info]

            # comments
            if "comment" in tde and not np.any(pd.isna(tde.comment)):
                if "schema_version" not in json:
                    json["schema_version"] = {}
                json["schema_version"]["comment"] = tde.comment.tolist()[0]

            # skip the photometry code if there is no photometry file
            # if there is a photometry file then we want to convert it below
            phot_sources = []
            if phot is not None:
                tde["obs_type"] = [
                    freq_to_obstype(vv * u.Unit(uu))
                    for vv, uu in zip(
                        tde.band_eff_freq.values,
                        tde.band_eff_freq_unit.values,
                    )
                ]

                unique_filter_keys = []
                index_for_match = []
                json["photometry"] = []

                if "telescope" in tde:
                    to_grpby = ["bibcode", "telescope", "obs_type"]
                else:
                    to_grpby = ["bibcode", "obs_type"]

                for grp_keys, p in tde.groupby(to_grpby, dropna=False):
                    if len(grp_keys) == 3:
                        src, tele, obstype = grp_keys
                    else:
                        src, obstype = grp_keys
                        tele = None

                    if src not in phot_sources:
                        phot_sources.append(src)

                    if len(np.unique(p.flux_unit)) == 1:
                        raw_units = p.flux_unit.tolist()[0]
                    else:
                        raw_units = p.flux_unit.tolist()

                    # add a column to phot with the unique filter key
                    if obstype == "radio":
                        filter_uq_key = (
                            p.band_eff_freq.astype(str)
                            + p.band_eff_freq_unit.astype(str)
                        ).tolist()

                    elif obstype in ("uvoir", "xray"):
                        filter_uq_key = p["filter"].astype(str).tolist()

                    else:
                        raise ValueError("not prepared for this obstype!")

                    unique_filter_keys += filter_uq_key
                    index_for_match += p.index.tolist()

                    if "upperlimit" not in p:
                        p["upperlimit"] = False

                    json_phot = dict(
                        reference=src,
                        raw=p.flux.astype(float).tolist(),
                        raw_err=p.flux_err.astype(float).tolist(),
                        raw_units=raw_units,
                        date=p.date.tolist(),
                        date_format=p.date_format.tolist(),
                        upperlimit=p.upperlimit.tolist(),
                        filter_key=filter_uq_key,
                        obs_type=obstype,
                    )

                    if not pd.isna(tele):
                        json_phot["telescope"] = tele

                    if pd.isna(tele) and obstype == "xray":
                        raise ValueError("The telescope is required for X-ray data!")

                    # check the minimum and maximum filter values
                    if obstype == "xray" and (
                        "filter_min" not in p or "filter_max" not in p
                    ):
                        raise ValueError(
                            "Minimum and maximum filters required for X-ray data!"
                        )

                    # check optional keys
                    optional_keys = [
                        "date_err",
                        "sigma",
                        "instrument",
                        "phot_type",
                        "exptime",
                        "aperature",
                        "observer",
                        "reducer",
                        "pipeline",
                    ]
                    for k in optional_keys:
                        if k in p and not np.all(pd.isna(p[k])):
                            # fill the nan values
                            # this is to match with the official json format
                            # and works with arangodb document structure
                            p[k].fillna("null", inplace=True)

                            json_phot[k] = p[k].tolist()

                    # handle more detailed uncertainty information
                    raw_err_detail = {}
                    for key in ["statistical_err", "systematic_err", "iss_err"]:
                        if key in p and not np.all(pd.isna(p[key])):
                            k = key.split("_")[0]

                            # fill the nan values
                            # this is to match with the official json format
                            # and works with arangodb document structure
                            p[key].fillna(0, inplace=True)

                            raw_err_detail[k] = p[key].tolist()

                    if len(raw_err_detail) > 0:
                        json_phot["raw_err_detail"] = raw_err_detail

                    # check the possible corrections
                    corrs = ["val_k", "val_s", "val_host", "val_av", "val_hostav"]
                    for c in corrs:
                        bool_v_key = c.replace("val", "corr")
                        json_phot[c] = False

                        if c in p:
                            # fill the nan values
                            # this is to match with the official json format
                            # and works with arangodb document structure
                            p[c].fillna("null", inplace=True)

                            json_phot[c] = p[c].tolist()
                            json_phot[bool_v_key] = [v != "null" for v in json_phot[c]]

                    json["photometry"].append(json_phot)

                tde["filter_uq_key"] = pd.Series(
                    unique_filter_keys, index=index_for_match
                )

                # filter alias
                # radio filters first
                filter_keys1 = ["filter_uq_key", "band_eff_wave", "band_eff_wave_unit"]
                if "filter_min" in tde:
                    filter_keys1.append("filter_min")
                if "filter_max" in tde:
                    filter_keys1.append("filter_max")

                filter_map = (
                    tde[filter_keys1].drop_duplicates().set_index("filter_uq_key")
                )  # .to_dict(orient='index')
                try:
                    filter_map_radio = filter_map.to_dict(orient="index")
                except Exception:
                    print(filter_map)
                    print(name)
                    raise Exception

                json["filter_alias"] = []
                for filt, val in filter_map_radio.items():
                    obs_type = freq_to_obstype(
                        float(val["band_eff_wave"]) * u.Unit(val["band_eff_wave_unit"])
                    )
                    if obs_type == "radio":
                        filter_name = freq_to_band(
                            (
                                float(val["band_eff_wave"])
                                * u.Unit(val["band_eff_wave_unit"])
                            ).to(u.GHz, equivalencies=u.spectral())
                        )
                    else:
                        filter_name = filt

                    filter_alias_dict = dict(
                        filter_key=filt,
                        filter_name=filter_name,
                        wave_eff=float(val["band_eff_wave"]),
                        wave_units=val["band_eff_wave_unit"],
                    )

                    if "filter_min" in val:
                        filter_alias_dict["wave_min"] = (
                            val["filter_min"] * u.Unit(phot.filter_eff_units)
                        ).to(
                            u.Unit(
                                val["band_eff_wave_unit"], equivalencies=u.spectral()
                            )
                        )

                    if "filter_max" in val:
                        filter_alias_dict["wave_max"] = (
                            val["filter_max"] * u.Unit(phot.filter_eff_units)
                        ).to(
                            u.Unit(
                                val["band_eff_wave_unit"], equivalencies=u.spectral()
                            )
                        )

                    json["filter_alias"].append(filter_alias_dict)

            # reference alias
            # gather all the bibcodes
            all_bibcodes = [tde.coord_bibcode[0]] + phot_sources
            if (
                "redshift_bibcode" in tde
                and tde.redshift_bibcode[0] not in all_bibcodes
                and not np.any(pd.isna(tde.redshift))
            ):
                all_bibcodes.append(tde.redshift_bibcode[0])

            if (
                "luminosity_distance_bibcode" in tde
                and tde.luminosity_distance_bibcode[0] not in all_bibcodes
                and not np.any(pd.isna(tde.luminosity_distance))
            ):
                all_bibcodes.append(tde.luminosity_distance_bibcode[0])

            if (
                "comoving_distance_bibcode" in tde
                and tde.comoving_distance_bibcode[0] not in all_bibcodes
                and not np.any(pd.isna(tde.comoving_distance))
            ):
                all_bibcodes.append(tde.comoving_distance_bibcode[0])

            if (
                "discovery_date_bibcode" in tde
                and tde.discovery_date_bibcode[0] not in all_bibcodes
                and not np.any(pd.isna(tde.discovery_date))
            ):
                all_bibcodes.append(tde.discovery_date_bibcode[0])

            if (
                "classification_bibcode" in tde
                and tde.classification_bibcode[0] not in all_bibcodes
                and not np.any(pd.isna(tde.classification))
            ):
                all_bibcodes.append(tde.classification_bibcode[0])

            if (
                "host_bibcode" in tde
                and tde.host_bibcode not in all_bibcodes
                and not np.any(pd.isna(tde.host_bibcode))
            ):
                all_bibcodes.append(tde.host_bibcode[0])

            # find the hrn's for all of these bibcodes
            uq_bibcodes, all_hrns = bibcode_to_hrn(all_bibcodes)

            # package these into the reference alias
            json["reference_alias"] = [
                dict(name=name, human_readable_name=hrn)
                for name, hrn in zip(uq_bibcodes, all_hrns)
            ]

            all_jsons.append(Transient(json))

        if db is None:
            db = Otter(datadir=local_outpath)
        else:
            db.DATADIR = local_outpath

        # always save this document as a new one
        db.save(all_jsons)
        db.generate_summary_table(save=True)
        return db



"""
Test the utility functions used throughout the OTTER API
"""

from copy import deepcopy
from otter import util


def test_filter_to_obstype():
    """
    Test the converion from a filtername to a obstype variable. This is used a lot
    during the data cleaning process
    """

    assert util.filter_to_obstype("FUV") == "uvoir"
    assert util.filter_to_obstype("r") == "uvoir"


def test_clean_schema():
    """
    Also used a lot during the data cleaning process. This function tests the
    clean_schema method.
    """

    subschema = deepcopy(util.distance_schema)

    subschema["value"] = "foo"
    subschema["reference"] = "bar"

    cleaned_schema = util.clean_schema(subschema)
    assert "unit" not in cleaned_schema
    assert "value" in cleaned_schema
    assert "reference" in cleaned_schema
    assert "error" not in cleaned_schema
    assert cleaned_schema["value"] == "foo"
    assert cleaned_schema["reference"] == "bar"


def test_bibcode_to_hrn():
    """
    Since GitHub can't have an ADS API token we cant really implement these tests
    until we figure it out...
    """
    pass



"""
Test the transient getter methods. I distinguish this from some of the other
methods (like the __add__ and corresponding private methods) since they
are more complex and require additional tests.
"""

import pytest
import numpy as np
from otter import Transient, Host
from otter.exceptions import OtterLimitationError
from astropy.coordinates import SkyCoord
from astropy.time import Time


def test_transient_constructor():
    """
    Test the constructor and make sure we have the expected instance variables
    """
    test_json = generate_test_json()

    # make sure the constructor works
    t = Transient(test_json)

    # test that the default_name attribute is correct
    assert t.default_name == "Sw J1644+57", "default_name is incorrect!"

    # test that the srcmap is correct
    # choose a few at random
    srcmap_msg = "Something with the srcmap is wrong!"
    assert t.srcmap["TNS"] == "TNS", srcmap_msg
    assert t.srcmap["2012MNRAS.421.1942W"] == "Wiersema et al. (2012)", srcmap_msg
    assert t.srcmap["2011ApJ...737..103S"] == "Schlafly & Finkbeiner (2011)", srcmap_msg


def test_getitem():
    """
    Test the __getitem__ method
    """
    msg = "Something is broken with __getitem__"

    # create a transient object to test
    t = Transient(generate_test_json())

    # test the "classic" dictionary syntax
    assert t["name"]["default_name"] == "Sw J1644+57", msg
    assert t["name"]["alias"][-1]["value"] == "Swift J1644+57", msg
    assert t["coordinate"][0]["ra"] == "16 44 49.93130", msg

    # use the same tests but with the "/" syntax
    assert t["name/default_name"] == "Sw J1644+57", msg
    assert t["name/alias"][-1]["value"] == "Swift J1644+57", msg
    assert t["coordinate"][0]["ra"] == "16 44 49.93130", msg

    # test it where we pass in a list or tuple
    test_keys = ["name", "photometry"]
    out = t[test_keys]
    assert out["name/default_name"] == "Sw J1644+57", msg
    assert out["photometry"][0]["raw"][0] == 0.25, msg


def test_setitem():
    """
    Test the __setitem__ method
    """
    msg = "somethings wrong with setitem!"

    t = Transient()

    # first test is basic
    test1 = {"mytest": "nothing to see here"}
    t["test1"] = test1

    assert t["test1"] == test1, msg
    assert t["test1"]["mytest"] == "nothing to see here", msg

    # next test tries to set an item with the / syntax
    t["test1/myothertest"] = "this is cool!"
    assert t["test1"]["myothertest"] == "this is cool!", msg
    assert t["test1/myothertest"] == "this is cool!", msg

    # now try overwriting mytest
    t["test1/mytest"] = "this is better"
    assert t["test1"]["mytest"] == "this is better", msg
    assert t["test1/mytest"] == "this is better", msg


def test_delitem():
    """
    Test the __delitem__ method
    """

    msg = "something went wrong with __delitem__ method!"

    # test with the classical syntax
    t = Transient({"test1": {"mytest": "nothing to see here"}})

    del t["test1"]["mytest"]
    assert t["test1"] == {}, msg

    # test with the '/' syntax
    # test with the classical syntax
    t = Transient({"test1": {"mytest": "nothing to see here"}})

    with pytest.raises(OtterLimitationError):
        del t["test1/mytest"]


def test_len():
    msg = "Something is wrong with the __len__ method!"

    t = Transient({"test1": {"mytest": "nothing to see here", "other": 1}})

    assert len(t) == 1, msg
    assert len(t["test1"]) == 2, msg


def test_repr():
    """
    Test the string representation of a Transient object
    """
    msg = "Something with the string representation is unexpected!"
    t = Transient(generate_test_json())

    str_repr = str(t)
    corr_str_repr = f"Transient(\n\tName: Sw J1644+57,\n\tKeys: {t.keys()}\n)"

    assert str_repr == corr_str_repr, msg


def test_iter():
    """
    Test that my __iter__ overwrite works
    """
    true_vals = [
        {"value": "Sw J1644+57", "reference": ["Swift"]},
        {"value": "GRB 110328A", "reference": [["2011Sci...333..203B"]]},
        {"value": "Swift J164449.3+573451", "reference": [["2011Sci...333..203B"]]},
        {
            "value": "Swift J1644+57",
            "reference": [["2017ApJ...838..149A", "2011Sci...333..203B"]],
        },
    ]
    to_iterate = Transient(generate_test_json())["name/alias"]
    for found, true in zip(to_iterate, true_vals):
        assert found == true, "iterating is not working!"


def test_keys():
    """
    Test the keys method
    """
    msg = "Something is wrong with the keys() method!"
    true_keys = [
        "schema_version",
        "distance",
        "filter_alias",
        "reference_alias",
        "date_reference",
        "name",
        "photometry",
        "coordinate",
        "classification",
        "host",
    ]

    t = Transient(generate_test_json())

    assert list(t.keys()) == true_keys, msg


def test_get_meta():
    """
    Test the get meta method of Transient
    """

    msg = "get_meta method is broken!"

    t = Transient(generate_test_json())
    meta = t.get_meta()
    meta2 = t.get_meta(keys=["name", "classification"])

    # check the keys are correct
    true_meta_keys = [
        "schema_version",
        "distance",
        "filter_alias",
        "reference_alias",
        "date_reference",
        "name",
        "coordinate",
        "classification",
        "host",
    ]
    assert list(meta.keys()) == true_meta_keys, msg
    assert list(meta2.keys()) == ["name", "classification"], msg

    # check that most of the values are correct
    # name first
    assert meta["name/default_name"] == "Sw J1644+57", msg
    assert meta["name/default_name"] == t.default_name, msg
    assert meta2["name/default_name"] == "Sw J1644+57", msg

    # now classification
    assert meta["classification"][0]["object_class"] == "TDE", msg
    assert meta2["classification"][0]["object_class"] == "TDE", msg

    # now distance
    assert meta["distance"][0]["value"] == "0.354", msg
    with pytest.raises(KeyError):
        meta2["distance"]  # make sure this key isn't in this metadata


def test_get_skycoord():
    """
    Test the get skycoord method
    """

    t = Transient(generate_test_json())

    skycoord = t.get_skycoord()

    assert isinstance(skycoord, SkyCoord), "get_skycoord did not return a SkyCoord!"
    assert str(skycoord.ra) == "251d12m28.9695s", "RA does not match!"
    assert str(skycoord.dec) == "57d34m59.6893s", "Dec does not match!"


def test_get_discovery_date():
    """
    test the get_discovery_date method
    """

    t = Transient(generate_test_json())
    discdate = t.get_discovery_date()

    assert isinstance(discdate, Time), "get_discovery_date did not return a Time!"
    assert str(discdate) == "2011-03-29 00:00:00.000", "The time string is incorrect!"


def test_get_redshift():
    """
    Test the get redshift method
    """

    t = Transient(generate_test_json())
    z = t.get_redshift()

    assert float(z) == 0.354, "get_redshift did not return the correct value!"


def test_get_host():
    """
    Test the get host method

    This is pretty bare bones because the rest of the output is tested with test_host
    """

    t = Transient(generate_test_json())

    host = t.get_host()
    assert isinstance(host, list)
    assert isinstance(host[0], Host)


def test_clean_photometry():
    """
    Make sure we can correctly clean the photometry
    """

    msg = "Something is broken with the photometry cleaning!"
    t = Transient(generate_test_json())

    # first with just the default options
    phot = t.clean_photometry()
    uq_obs_types = phot["obs_type"].unique()

    assert all(ot in uq_obs_types for ot in ["radio", "uvoir", "xray"]), msg
    assert len(phot["converted_flux_unit"].unique()) == 1, msg
    assert phot["converted_flux_unit"][0] == "mag(AB)", msg

    uvoir = phot[phot.obs_type == "uvoir"]
    assert np.isclose(phot["converted_flux"].iloc[0], 17.905, atol=1e-2), msg
    assert 17.59 in list(uvoir["converted_flux"]), msg

    # then with different returned units
    # and only get the radio data
    phot_non_default = t.clean_photometry(
        flux_unit="uJy", freq_unit="MHz", obs_type="radio"
    ).reset_index()

    assert len(phot_non_default.obs_type.unique()) == 1, msg
    assert phot_non_default.obs_type.iloc[0] == "radio", msg
    assert np.isclose(phot_non_default.converted_flux.iloc[0], 0.25e3), msg


# a test json file
# I use this here instead of reading in existing files
# because I want to be able to control exactly what goes
# into the test cases!
def generate_test_json():
    """
    Generate a test json file

    This is a mixture of data on Sw J1644 plus some random
    other photometry points I put in from AT2018zr
    """

    test_json = {
        "schema_version": {"value": "0", "comment": "Copied from tde.space"},
        "distance": [
            {
                "value": "0.354",
                "reference": [
                    "2011Natur.476..425Z",
                    "2012ApJ...748...36B",
                    "2012MNRAS.421.1942W",
                    "2013ApJ...767..152Z",
                    "2016MNRAS.462L..66Y",
                    "2018ApJ...854...86E",
                ],
                "computed": False,
                "default": True,
                "distance_type": "redshift",
            },
            {
                "value": "0.3543",
                "reference": ["2017ApJ...838..149A", "2011Sci...333..203B"],
                "computed": False,
                "default": True,
                "distance_type": "redshift",
            },
            {
                "value": "1942",
                "reference": [
                    "2017ApJ...838..149A",
                    "2016A&A...594A..13P",
                    "2011Sci...333..203B",
                    "2017ApJ...835...64G",
                ],
                "computed": False,
                "default": True,
                "distance_type": "luminosity",
                "unit": "Mpc",
            },
            {
                "value": "1434",
                "reference": [
                    "2017ApJ...838..149A",
                    "2016A&A...594A..13P",
                    "2011Sci...333..203B",
                    "2017ApJ...835...64G",
                ],
                "computed": False,
                "default": True,
                "distance_type": "comoving",
                "unit": "Mpc",
            },
        ],
        "filter_alias": [
            {"filter_key": "4.9GHz", "freq_eff": 4.9, "freq_units": "GHz"},
            {
                "filter_key": "0.3 - 2.0",
                "wave_eff": 0.7293188143129428,
                "wave_min": 0.6199209921660013,
                "wave_max": 4.132806614440009,
                "wave_units": "nm",
            },
            {"filter_key": "g", "wave_eff": 1, "wave_units": "nm"},
            {"filter_key": "r", "wave_eff": 1, "wave_units": "nm"},
        ],
        "reference_alias": [
            {
                "name": "2011Natur.476..425Z",
                "human_readable_name": "Zauderer et al. (2011)",
            },
            {
                "name": "2012ApJ...748...36B",
                "human_readable_name": "Berger et al. (2012)",
            },
            {
                "name": "2012MNRAS.421.1942W",
                "human_readable_name": "Wiersema et al. (2012)",
            },
            {
                "name": "2013ApJ...767..152Z",
                "human_readable_name": "Zauderer et al. (2013)",
            },
            {
                "name": "2016MNRAS.462L..66Y",
                "human_readable_name": "Yang et al. (2016)",
            },
            {
                "name": "2018ApJ...854...86E",
                "human_readable_name": "Eftekhari et al. (2018)",
            },
            {
                "name": "2017ApJ...838..149A",
                "human_readable_name": "Auchettl, Guillochon, & Ramirez-Ruiz (2017)",
            },
            {
                "name": "2016A&A...594A..13P",
                "human_readable_name": "Planck Collaboration et al. (2016)",
            },
            {
                "name": "2011Sci...333..203B",
                "human_readable_name": "Bloom et al. (2011)",
            },
            {
                "name": "2011ApJ...737..103S",
                "human_readable_name": "Schlafly & Finkbeiner (2011)",
            },
            {
                "name": "2017ApJ...835...64G",
                "human_readable_name": "Guillochon et al. (2017)",
            },
        ],
        "date_reference": [
            {
                "value": 55648.54,
                "date_format": "mjd",
                "reference": [
                    "2011Natur.476..425Z",
                    "2012ApJ...748...36B",
                    "2012MNRAS.421.1942W",
                    "2013ApJ...767..152Z",
                    "2016MNRAS.462L..66Y",
                    "2018ApJ...854...86E",
                ],
                "computed": False,
                "date_type": "discovery",
            },
            {
                "value": "2011-03-29",
                "date_format": "iso",
                "reference": ["2017ApJ...838..149A", "2017ApJ...835...64G"],
                "computed": False,
                "date_type": "discovery",
                "measurement_type": "discovery",
                "default": True,
            },
        ],
        "name": {
            "default_name": "Sw J1644+57",
            "alias": [
                {"value": "Sw J1644+57", "reference": ["Swift"]},
                {"value": "GRB 110328A", "reference": [["2011Sci...333..203B"]]},
                {
                    "value": "Swift J164449.3+573451",
                    "reference": [["2011Sci...333..203B"]],
                },
                {
                    "value": "Swift J1644+57",
                    "reference": [["2017ApJ...838..149A", "2011Sci...333..203B"]],
                },
            ],
        },
        "photometry": [
            {
                "reference": [
                    "2011Natur.476..425Z",
                    "2012ApJ...748...36B",
                    "2012MNRAS.421.1942W",
                    "2013ApJ...767..152Z",
                    "2016MNRAS.462L..66Y",
                    "2018ApJ...854...86E",
                ],
                "raw": [
                    0.25,
                    0.34,
                    0.34,
                    0.61,
                    0.82,
                    1.48,
                    1.47,
                    1.8,
                    2.1,
                    4.62,
                    4.84,
                    5.86,
                    9.06,
                    9.1,
                    9.1,
                    11.71,
                    12.93,
                    12.83,
                    13.29,
                    12.43,
                    12.17,
                    12.05,
                    12.24,
                    11.12,
                    8.9,
                    8.24,
                    8.63,
                    6.23,
                    4.21,
                    3.52,
                    2.34,
                    1.47,
                ],
                "raw_units": [
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                    "mJy",
                ],
                "date": [
                    55652.41,
                    55653.3,
                    55653.54,
                    55654.33,
                    55655.32,
                    55656.31,
                    55658.33,
                    55663.520000000004,
                    55671.32,
                    55684.4,
                    55699.19,
                    55716.15,
                    55743.18,
                    55760.16,
                    55775.05,
                    55792.16,
                    55812.92,
                    55823.01,
                    55845.950000000004,
                    55861.86,
                    55893.770000000004,
                    55951.49,
                    56032.46,
                    56102.200000000004,
                    56230.85,
                    56293.54,
                    56299.64,
                    56436.14,
                    56680.54,
                    56753.54,
                    57021.54,
                    57542.54,
                ],
                "date_format": [
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                    "mjd",
                ],
                "filter_key": [
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                    "4.9GHz",
                ],
                "computed": [
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
                "obs_type": [
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                    "radio",
                ],
                "upperlimit": [
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
                "corr_k": False,
                "corr_s": False,
                "corr_av": False,
                "corr_host": False,
                "corr_hostav": False,
            },
            {
                "raw": ["0.000427131", "0.000399543"],
                "raw_units": ["ct", "ct"],
                "value": ["3.05E-15", "2.86E-15"],
                "value_units": ["ergs/cm^2/s", "ergs/cm^2/s"],
                "filter_key": ["0.3 - 2.0", "0.3 - 2.0"],
                "obs_type": ["xray", "xray"],
                "date": ["56257.4", "57070.2"],
                "date_format": "MJD",
                "upperlimit": [True, True],
                "telescope": "Chandra",
                "corr_k": None,
                "corr_av": None,
                "corr_host": None,
                "corr_hostav": None,
                "reference": ["2017ApJ...838..149A"],
                "corr_s": None,
            },
            {
                "raw": ["0.0308041", "0.0993732", "0.0450751", "0.043094"],
                "raw_units": ["ct", "ct", "ct", "ct"],
                "value": ["8.75E-14", "1.66E-13", "7.53E-14", "7.20E-14"],
                "value_units": [
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                ],
                "filter_key": ["0.3 - 2.0", "0.3 - 2.0", "0.3 - 2.0", "0.3 - 2.0"],
                "obs_type": ["xray", "xray", "xray", "xray"],
                "date": ["48000", "48094.1", "48370", "48715.6"],
                "date_format": "MJD",
                "upperlimit": [True, True, True, True],
                "telescope": "ROSAT",
                "corr_k": None,
                "corr_av": None,
                "corr_host": None,
                "corr_hostav": None,
                "reference": ["2017ApJ...838..149A"],
                "corr_s": None,
            },
            {
                "raw": [
                    "1.0482",
                    "0.6817",
                    "0.101597",
                    "0.0362351",
                    "0.0225792",
                    "0.0122362",
                    "0.00300771",
                    "0.00126983",
                    "0.00156039",
                    "0.00173297",
                    "1.95E-03",
                    "1.86E-03",
                    "2.09E-03",
                    "1.97E-03",
                    "0.0019953",
                    "0.00171713",
                    "0.00184643",
                    "0.00252298",
                    "0.00236432",
                ],
                "raw_units": [
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                ],
                "value": [
                    "8.74E-12",
                    "5.53E-12",
                    "8.91E-13",
                    "3.04E-13",
                    "2.27E-13",
                    "9.46E-14",
                    "2.26E-14",
                    "2.21E-14",
                    "2.71E-14",
                    "3.00E-14",
                    "3.39E-14",
                    "3.23E-14",
                    "3.63E-14",
                    "3.42E-14",
                    "3.47E-14",
                    "2.99E-14",
                    "3.21E-14",
                    "4.38E-14",
                    "4.10E-14",
                ],
                "value_units": [
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                ],
                "filter_key": [
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                ],
                "obs_type": [
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                ],
                "date": [
                    "55649.6",
                    "55675.6",
                    "55749.3",
                    "55850.3",
                    "55947.6",
                    "56048.1",
                    "56153.6",
                    "56242.9",
                    "56335.7",
                    "56457",
                    "56551.70",
                    "56649.30",
                    "56751.40",
                    "56848.30",
                    "56945.1",
                    "57049",
                    "57151.3",
                    "57253.9",
                    "57318.4",
                ],
                "date_format": "MJD",
                "upperlimit": [
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
                "telescope": "Swift",
                "corr_k": None,
                "corr_av": None,
                "corr_host": None,
                "corr_hostav": None,
                "reference": ["2017ApJ...838..149A"],
                "corr_s": None,
            },
            {
                "raw": [
                    "0.0319464",
                    "11.3004",
                    "7.93976",
                    "0.928866",
                    "17.5",
                    "0.241757",
                    "1.72019",
                    "2.83",
                    "1.45692",
                    "1.51279",
                    "0.348197",
                    "0.614621",
                    "0.860492",
                    "0.0138642",
                    "1.08E-02",
                    "1.8",
                    "5.35E-01",
                ],
                "raw_units": [
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                    "ct",
                ],
                "value": [
                    "1.89E-14",
                    "9.03E-12",
                    "5.14E-12",
                    "6.93E-13",
                    "3.01E-11",
                    "1.53E-13",
                    "7.87E-13",
                    "4.86E-12",
                    "1.19E-13",
                    "4.07E-13",
                    "1.46E-13",
                    "3.64E-13",
                    "5.08E-13",
                    "8.22E-15",
                    "6.38E-15",
                    "3.09E-12",
                    "9.19E-13",
                ],
                "value_units": [
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                    "ergs/cm^2/s",
                ],
                "filter_key": [
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                    "0.3 - 2.0",
                ],
                "obs_type": [
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                    "xray",
                ],
                "date": [
                    "55651.4",
                    "55667.4",
                    "55681.3",
                    "55697.3",
                    "55697.6",
                    "55745.3",
                    "55757.1",
                    "55757.4",
                    "55769.1",
                    "55787.1",
                    "55801",
                    "55811",
                    "55836.9",
                    "56197.8",
                    "56205.8",
                    "56296.3",
                    "57231.4",
                ],
                "date_format": "MJD",
                "upperlimit": [
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
                "telescope": "XMM",
                "corr_k": None,
                "corr_av": None,
                "corr_host": None,
                "corr_hostav": None,
                "reference": ["2017ApJ...838..149A"],
                "corr_s": None,
            },
            {
                "raw": [
                    "21.44",
                    "21.13",
                    "21.65",
                    "21.06",
                    "20.85",
                    "20.47",
                    "18.85",
                    "18.84",
                    "18.77",
                    "18.64",
                    "18.64",
                    "18.16",
                    "17.98",
                    "17.59",
                    "17.59",
                ],
                "raw_units": "mag(AB)",
                "filter_key": [
                    "r",
                    "r",
                    "r",
                    "r",
                    "r",
                    "r",
                    "g",
                    "g",
                    "g",
                    "g",
                    "g",
                    "r",
                    "r",
                    "r",
                    "r",
                ],
                "obs_type": [
                    "uvoir",
                    "uvoir",
                    "uvoir",
                    "uvoir",
                    "uvoir",
                    "uvoir",
                    "uvoir",
                    "uvoir",
                    "uvoir",
                    "uvoir",
                    "uvoir",
                    "uvoir",
                    "uvoir",
                    "uvoir",
                    "uvoir",
                ],
                "date": [
                    "58101.360",
                    "58105.310",
                    "58154.220",
                    "58156.230",
                    "58158.230",
                    "58160.230",
                    "58166.260",
                    "58167.180",
                    "58168.164",
                    "58168.176",
                    "58170.392",
                    "58182.190",
                    "58183.180",
                    "58190.154",
                    "58190.155",
                ],
                "date_format": "MJD",
                "upperlimit": [
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
                "telescope": "P48",
                "corr_k": None,
                "corr_av": None,
                "corr_host": None,
                "corr_hostav": None,
                "reference": ["2018arXiv180902608V"],
                "corr_s": None,
            },
        ],
        "coordinate": [
            {
                "ra": "16 44 49.93130",
                "dec": "+57 34 59.6893",
                "epoch": "J2000",
                "system": "ICRS",
                "ra_units": "hourangle",
                "dec_units": "deg",
                "reference": [
                    "2011Natur.476..425Z",
                    "2012ApJ...748...36B",
                    "2012MNRAS.421.1942W",
                    "2013ApJ...767..152Z",
                    "2016MNRAS.462L..66Y",
                    "2018ApJ...854...86E",
                ],
                "computed": False,
                "default": True,
                "uuid": "b98cdbe9-fff7-415b-a6d6-78cbd88544b8",
                "coordinate_type": "equatorial",
            },
            {
                "reference": "b98cdbe9-fff7-415b-a6d6-78cbd88544b8",
                "computed": True,
                "coordinate_type": "galactic",
                "l": 86.71123273634593,
                "b": 39.44122227143398,
                "l_units": "deg",
                "b_units": "deg",
            },
            {
                "ra": "16:44:49.92",
                "dec": "+57:34:58.8",
                "ra_units": "hour",
                "dec_units": "deg",
                "coordinate_type": "equitorial",
            },
        ],
        "classification": [
            {
                "object_class": "TDE",
                "confidence": 1.0,
                "reference": [
                    "2011Natur.476..425Z",
                    "2011Sci...333..203B",
                    "2012ApJ...748...36B",
                    "2012MNRAS.421.1942W",
                    "2013ApJ...767..152Z",
                    "2016MNRAS.462L..66Y",
                    "2018ApJ...854...86E",
                ],
                "default": True,
            }
        ],
        "host": [
            {
                "host_name": "Swift J164449.3+573451",
                "host_ra": 251.20416666666665,
                "host_dec": 57.58083333333334,
                "host_ra_units": "deg",
                "host_dec_units": "deg",
                "reference": [
                    "2023PASP..135c4101G",
                    "2011Sci...333..199L",
                    "2011Sci...333..203B",
                    "2017ApJ...838..149A",
                ],
            }
        ],
    }

    return test_json



"""
Some utilities to create common plots for transients that use the OtterPlotter
"""

from __future__ import annotations
from warnings import warn

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from .otter_plotter import OtterPlotter
from ..exceptions import FailedQueryError
from ..io.otter import Transient, Otter


def query_quick_view(
    db: Otter = None,
    otter_path: str = None,
    ptype: str = "both",
    sed_dim: str = "freq",
    dt_over_t: float = 0,
    plotting_kwargs: dict = {},
    phot_cleaning_kwargs: dict = {},
    result_length_tol=10,
    **kwargs,
) -> list[plt.Figure]:
    """
    Queries otter and then plots all of the transients. It will either query the Otter
    object provided in db or construct an Otter object from otter_path.

    Args:
        db (otter.Otter) : The otter object to query
        otter_path (str) : The path to construct an otter path from
        ptype (str) : The plot type to generate. Valid options are
                      - both -> Plot both light curve and sed (default)
                      - sed -> Plot just the sed
                      - lc -> Plot just the light curve
        sed_dim (str) : The x dimension to plot in the SED. Options are "freq" or
                        "wave". Default is "freq".
        time_tol (float) : The tolerance to split the days by. Default is 1 day. must be
                           in units of days.
        plotting_kwargs (dict) : dictionary of key word arguments to pass to
                                 otter.plotter.plot_light_curve or
                                 otter.plotter.plot_sed.
        phot_cleaning_kwargs (dict) : Keyword arguments passed to
                                      otter.Transient.clean_photometry
        result_length_tol (int) : If the query result is longer than this it will throw
                                  an erorr to prevent 100s of plots from spitting out
                                  (and likely crashing your computer). Default is 10 but
                                  can be adjusted.
        **kwargs : Arguments to pass to otter.Otter.query

    Returns:
        A list of matplotlib pyplot Figure objects that we plotted

    """
    if db is None:
        if otter_path is not None:
            db = Otter(otter_path)
        else:
            raise ValueError("Either the db or otter_path arguments must be provided!")

    res = db.query(**kwargs)

    if len(res) > result_length_tol:
        raise RuntimeError(
            f"This query returned {len(res)} results which is greater than the given "
            + f"tolerance of {result_length_tol}! Either increase the result_length_tol"
            + " keyword or pass in a stricter query!"
        )

    figs = []
    for t in res:
        try:
            fig = quick_view(
                t, ptype, sed_dim, dt_over_t, plotting_kwargs, **phot_cleaning_kwargs
            )
        except (KeyError, FailedQueryError):
            warn(f"No photometry associated with {t.default_name}, skipping!")
            continue

        fig.suptitle(t.default_name)
        figs.append(fig)

    return figs


def quick_view(
    t: Transient,
    ptype: str = "both",
    sed_dim: str = "freq",
    dt_over_t: float = 0,
    plotting_kwargs: dict = {},
    **kwargs,
) -> plt.Figure:
    """
    Generate a quick view (not necessarily publication ready) of the transients light
    curve, SED, or both. Default is to do both.

    Args:
        t (otter.Transient) : An otter Transient object to grab photometry from
        ptype (str) : The plot type to generate. Valid options are
                      - both -> Plot both light curve and sed (default)
                      - sed -> Plot just the sed
                      - lc -> Plot just the light curve
        sed_dim (str) : The x dimension to plot in the SED. Options are "freq" or
                        "wave". Default is "freq".
        dt_over_t (float) : The tolerance to split the days by. Default is 1 day. must
                           be unitless.
        plotting_kwargs (dict) : dictionary of key word arguments to pass to
                                 otter.plotter.plot_light_curve or
                                 otter.plotter.plot_sed.
        **kwargs : Any other arguments to pass to otter.Transient.clean_photometry

    Returns:
        The matplotlib figure used for plotting.
    """
    backend = plotting_kwargs.get("backend", "matplotlib.pyplot")
    if backend not in {"matplotlib.pyplot", "matplotlib"}:
        raise ValueError(
            "Only matplotlib.pyplot backend is available for quick_view!"
            + " To use plotly, use the plotting functionality individually!"
        )

    allphot = t.clean_photometry(**kwargs)
    allphot = allphot.sort_values("converted_date")
    allphot["time_tol"] = dt_over_t * allphot["converted_date"]
    allphot["time_diff"] = allphot["converted_date"].diff().fillna(-np.inf)
    allphot["time_grp"] = (allphot.time_diff > allphot.time_tol).cumsum()

    plt_lc = (ptype == "both") or (ptype == "lc")
    plt_sed = (ptype == "both") or (ptype == "sed")

    if ptype == "both":
        fig, (lc_ax, sed_ax) = plt.subplots(1, 2)
    elif ptype == "sed":
        fig, sed_ax = plt.subplots()
    elif ptype == "lc":
        fig, lc_ax = plt.subplots()

    if np.all(pd.isna(allphot.converted_flux_err)):
        flux_err = None
    else:
        flux_err = allphot.converted_flux_err

    if plt_lc:
        for filt, phot in allphot.groupby("filter_name"):
            plot_light_curve(
                date=phot.converted_date,
                flux=phot.converted_flux,
                flux_err=flux_err[allphot.filter_name == filt],
                xlabel=f"Date [{phot.converted_date_unit.values[0]}]",
                ylabel=f"Flux [{phot.converted_flux_unit.values[0]}]",
                ax=lc_ax,
                label=filt,
                **plotting_kwargs,
            )

    if plt_sed:
        for grp_name, phot in allphot.groupby("time_grp"):
            if sed_dim == "wave":
                wave_or_freq = phot.converted_wave
                xlab = f"Wavelength [{phot.converted_wave_unit.values[0]}]"
            elif sed_dim == "freq":
                wave_or_freq = phot.converted_freq
                xlab = f"Frequency [{phot.converted_freq_unit.values[0]}]"
            else:
                raise ValueError("sed_dim value is not recognized!")

            plot_sed(
                wave_or_freq=wave_or_freq,
                flux=phot.converted_flux,
                flux_err=flux_err[allphot.time_grp == grp_name],
                ax=sed_ax,
                xlabel=xlab,
                ylabel=f"Flux [{phot.converted_flux_unit.values[0]}]",
                label=phot.converted_date.mean(),
                **plotting_kwargs,
            )

        sed_ax.set_xscale("log")

    return fig


def plot_light_curve(
    date: float,
    flux: float,
    date_err: float = None,
    flux_err: float = None,
    fig=None,
    ax=None,
    backend: str = "matplotlib",
    xlabel: str = "Date",
    ylabel: str = "Flux",
    **kwargs,
):
    """
    Plot the light curve for the input data

    Args:
        date (float): MJD dates
        flux (float): Flux
        date_err (float): optional error on the MJD dates
        flux_err (float): optional error on the flux
        fig (float): matplotlib fig object, optional. Will be created if not provided.
        ax (float): matplitlib axis object, optional. Will be created if not provided.
        backend (str): backend for plotting. options: "matplotlib" (default) or "plotly"
        xlabel (str): x-axis label
        ylabel (str): y-axis label
        **kwargs: keyword arguments to pass to either plotly.graph_objects.add_scatter
                  or matplotlib.pyplot.errorbar

    Returns:
       Either a matplotlib axis or plotly figure
    """

    plt = OtterPlotter(backend)
    fig = plt.plot(date, flux, date_err, flux_err, ax=ax, **kwargs)

    if backend == "matplotlib":
        fig.set_ylabel(ylabel)
        fig.set_xlabel(xlabel)

    elif backend == "plotly":
        fig.update_layout(xaxis_title=xlabel, yaxis_title=ylabel)

    return fig


def plot_sed(
    wave_or_freq: float,
    flux: float,
    wave_or_freq_err: float = None,
    flux_err: float = None,
    fig=None,
    ax=None,
    backend: str = "matplotlib",
    xlabel: str = "Frequency or Wavelength",
    ylabel: str = "Flux",
    **kwargs,
):
    """
    Plot the SED for the input data

    Args:
        wave_or_freq (float): wave or frequency array
        flux (float): Flux
        wave_or_freq_err (float): optional error on the MJD dates
        flux_err (float): optional error on the flux
        fig (float): matplotlib fig object, optional. Will be created if not provided.
        ax (float): matplitlib axis object, optional. Will be created if not provided.
        backend (str): backend for plotting. Options: "matplotlib" (default) or "plotly"
        xlabel (str): x-axis label
        ylabel (str): y-axis label
        **kwargs: keyword arguments to pass to either plotly.graph_objects.add_scatter
                  or matplotlib.pyplot.errorbar

    Returns:
       Either a matplotlib axis or plotly figure
    """

    plt = OtterPlotter(backend)
    fig = plt.plot(wave_or_freq, flux, wave_or_freq_err, flux_err, ax=ax, **kwargs)

    if backend == "matplotlib":
        fig.set_ylabel(ylabel)
        fig.set_xlabel(xlabel)

    elif backend == "plotly":
        fig.update_layout(xaxis_title=xlabel, yaxis_title=ylabel)

    return fig



"""
Test the OTTER DataFinder class
"""

import pytest
import pandas as pd
from otter import DataFinder
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astroquery.utils import TableList
from requests.exceptions import ConnectionError, Timeout


def construct_data_finder(**kwargs):
    """
    Construct a DataFinder object and return it

    Based on the ASASSN-14li coordinates
    """
    ra = 192.06347
    dec = 17.77402
    return DataFinder(ra, dec, "deg", "deg", **kwargs)


def test_data_finder_constructor():
    """
    Test the constructor and make sure the instance variables are set correctly
    """
    # first, very basic, test
    df1 = construct_data_finder()
    assert isinstance(df1.coord, SkyCoord), "Not constructing a SkyCoord"
    assert df1.coord.ra.value == 192.06347, "Incorrect RA"
    assert df1.coord.dec.value == 17.77402, "Incorrect Dec"
    assert df1.name is None
    assert df1.z is None
    assert df1.redshift is None
    assert df1.bibcodes is None

    # next, more complex test
    df2 = construct_data_finder(
        name="ASASSN-14li",
        reference=["my_fake_reference"],
        redshift=0.1,  # this is made up for testing
    )
    assert isinstance(df1.coord, SkyCoord), "Not constructing a SkyCoord"
    assert df2.coord.ra.value == 192.06347, "Incorrect RA"
    assert df2.coord.dec.value == 17.77402, "Incorrect Dec"
    assert df2.name == "ASASSN-14li"
    assert df2.z == 0.1
    assert df2.redshift == 0.1
    assert df2.bibcodes == ["my_fake_reference"]


def test_repr():
    """
    Make sure the string representation is correct
    """
    df1 = construct_data_finder()
    assert str(df1) == "No Name DataFinder @ (RA, Dec)=(192.06347 deg,17.77402 deg)"

    df2 = construct_data_finder(
        name="ASASSN-14li",
        reference=["my_fake_reference"],
        redshift=0.1,  # this is made up for testing
    )
    assert str(df2) == "ASASSN-14li @ (RA, Dec)=(192.06347 deg,17.77402 deg)"


def test_iter():
    """
    Test the __iter__ method
    """
    keys1 = {"host_ra", "host_dec", "host_ra_units", "host_dec_units"}
    df1 = construct_data_finder()
    assert len(list({k for k, _ in df1} - keys1)) == 0

    keys2 = {
        "host_ra",
        "host_dec",
        "host_ra_units",
        "host_dec_units",
        "host_name",
        "host_redshift",
        "reference",
    }
    df2 = construct_data_finder(
        name="ASASSN-14li",
        reference=["my_fake_reference"],
        redshift=0.1,  # this is made up for testing
    )
    assert len(list({k for k, _ in df2} - keys2)) == 0


def test_query_simbad():
    """
    Test the query_simbad method
    """
    df1 = construct_data_finder()
    res = df1.query_simbad()

    assert len(res) == 3


def test_query_vizier():
    """
    Test the query_vizier method
    """
    df1 = construct_data_finder()
    res = df1.query_vizier()

    assert isinstance(res, TableList)
    assert len(res) >= 20, "missing some data"
    assert "II/294/sdss7" in res.keys()


def broken_test_query_atlas():
    """
    Test the query_atlas method

    THIS IS BROKEN ON GITHUB ACTIONS BECAUSE WE CANT SET THE ENV VARS!!!

    This one will inherently take a while!!! But it's because of atlas not us...
    """
    df1 = construct_data_finder()
    res = df1.query_atlas()

    assert "mjd" in res.columns
    assert "uJy" in res.columns
    assert "duJy" in res.columns
    assert len(res) >= 96, "missing some data!"


def test_query_ptf():
    """
    Test the method that queries palomer transient facility
    """
    df1 = construct_data_finder()
    res = df1.query_ptf()

    assert isinstance(res, Table)
    assert "obsmjd" in res.columns
    assert "mag_autocorr" in res.columns
    assert "meanmag" in res.columns
    assert len(res) >= 32, "Missing some PTF data!"


def test_query_ztf():
    """
    Test the method that queries the zwicky transient facility
    """
    df1 = construct_data_finder()
    res = df1.query_ztf()

    assert isinstance(res, Table)
    assert "mjd" in res.columns
    assert "mag" in res.columns
    assert "ra" in res.columns
    assert len(res) >= 1296, "Missing some ZTF data!"


def test_query_asassn():
    """
    Test the method that queries the ASASSN survey
    """
    df1 = construct_data_finder()
    res = df1.query_asassn()

    assert isinstance(res, pd.DataFrame)
    assert "asas_sn_id" in res.columns
    assert "jd" in res.columns
    assert "flux" in res.columns
    assert len(res) >= 1158, "Missing some ZTF data!"


def test_query_wise():
    """
    Test querying wise
    """
    df1 = construct_data_finder()
    res = df1.query_wise()

    assert isinstance(res, pd.DataFrame)
    assert "flux" in res.columns
    assert "date_mjd" in res.columns
    assert "name" in res.columns
    assert len(res) >= 35, "Missing some WISE data!"


@pytest.mark.skip(
    reason="There is a known issue with setuptools that makes this break\
when run in most conda environments!"
)
def test_query_alma():
    """
    Test querying ALMA
    """
    df1 = construct_data_finder()
    res = df1.query_alma()

    assert isinstance(res, Table)
    assert "obs_id" in res.columns
    assert "access_url" in res.columns


def test_query_first():
    """
    Test queryiung the FIRST radio survey
    """
    df1 = construct_data_finder()
    res = df1.query_first()

    assert isinstance(res, TableList)
    assert len(res) >= 1
    assert "J/ApJ/737/45/table2" in res.keys()


def test_query_nvss():
    """
    Test querying the NVSS radio survey
    """
    df1 = construct_data_finder()
    res = df1.query_nvss()

    assert isinstance(res, TableList)
    assert len(res) >= 1
    assert "J/ApJ/737/45/table1" in res.keys()


@pytest.mark.skip(
    reason="""This has a tendancy to timeout, not sure why but it sounds like its
    not on us (hopefully)..."""
)
def test_query_sparcl():
    """
    Test querying SPARCL for spectra

    Note: Sparcl likes to throttle us so I'm wrapping this in a try/except for some very
    specific errors that requests will throw
    """

    df1 = construct_data_finder()
    try:
        res = df1.query_sparcl()
        assert isinstance(res, Table)
        assert len(res) >= 1, "Missing some spectroscopic data!"

    except (ConnectionError, Timeout):  # NOTE: NEVER add AttributeError to this list
        pass


@pytest.mark.skip(reason="pyvo is rejecting our queries of the heasarc catalog")
def test_query_heasarc():
    """
    Test querying the Heasarc x-ray catalog
    """

    df1 = construct_data_finder()

    # test with x-ray
    res = df1.query_heasarc(catalog="xray")
    assert isinstance(res, Table)
    assert len(res) >= 14, "Missing some HEASARC data"

    # test with radio
    res2 = df1.query_heasarc(catalog="radio")
    assert isinstance(res2, Table)
    assert len(res2) >= 2, "Missing some HEASARC data"



"""
OtterPlotter Class which handles the backend of the plotting.

Currently supported backends are:
- matplotlib
- plotly
"""

from __future__ import annotations
import importlib


class OtterPlotter:
    """
    Handles the backend for the "plotter" module

    Args:
        backend (string): a string of the module name to import and use
                          as the backend. Currently supported are "matplotlib",
                          "matplotlib.pyplot", "plotly", and "plotly.graph_objects"
    """

    def __init__(self, backend):
        if backend == "matplotlib.pyplot":
            self.backend = backend
        elif backend == "pyplot.graph_objects":
            self.backend = backend
        elif "plotly" in backend and "graph_objects" not in backend:
            self.backend = "plotly.graph_objects"
        elif "matplotlib" in backend and "pyplot" not in backend:
            self.backend = "matplotlib.pyplot"
        else:
            raise ValueError("Not a valid backend string!")

        self.plotter = importlib.import_module(self.backend)

        if self.backend == "matplotlib.pyplot":
            self.plot = self._plot_matplotlib
        elif self.backend == "plotly.graph_objects":
            self.plot = self._plot_plotly
        else:
            raise ValueError("Unknown plotting backend!")

    def _plot_matplotlib(self, x, y, xerr=None, yerr=None, ax=None, **kwargs):
        """
        General plots using matplotlib, is called by _matplotlib_light_curve and
        _matplotlib_sed
        """

        if ax is None:
            _, ax = self.plotter.subplots()

        ax.errorbar(x, y, xerr=xerr, yerr=yerr, **kwargs)
        return ax

    def _plot_plotly(self, x, y, xerr=None, yerr=None, ax=None, *args, **kwargs):
        """
        General plotting method using plotly, is called by _plotly_light_curve and
        _plotly_sed
        """

        if ax is None:
            go = self.plotter.Figure()
        else:
            go = ax

        fig = go.add_scatter(
            x=x, y=y, error_x=dict(array=xerr), error_y=dict(array=yerr), **kwargs
        )

        return fig
